/** @file
/////////////////////////////////////////////////////////////////////
//
// Verified Systems International GmbH
// Parkstrasse 78 
// D-28209 Bremen
// Germany
// Tel.  : +49 421 344576
// Fax   : +49 421 344586
// e-mail: info@verified.de
//
//--------------------------------------------------------------------
//
// (C) Copyright Verified Systems International GmbH 
//     $Date: 2007/08/21 08:02:04 $
//
//--------------------------------------------------------------------
//
// Product: RT-Tester template for test startup operations
//
//--------------------------------------------------------------------
//
// File Identification: 
//
// $RCSfile: RTTstartup.c,v $
//
// $Header: /home/repository/VVTCVS/CVS/rtt-swi/tpl/RTTstartup.c,v 1.47 2007/08/21 08:02:04 oli Exp $
//
// $Revision: 1.47 $
//
// First edition by: Markus Dahlweid
// Last update by    $Author: oli $
//
//--------------------------------------------------------------------
//
// Description: 
//   
//
//--------------------------------------------------------------------*/
/** startup file for RT-Tester */ 

/************************************************************
  Includes
*************************************************************/

#include <stdio.h> 
#include <signal.h> 
#ifndef _CADUL
#include <unistd.h>
#endif

#include "rtt_test_config.h"
#include "rtt_test_types.h"

#include "rttservicelib.h"

#include "RTTcontext.h"
#include "RTTschedule.h"

#ifdef RTT_CLUSTER
#include "rttclusterdlib.h"
#endif

/** RTTschedule.c */
#if ( (defined(_LINUX) || defined(RTT_PTHREAD)) && ! defined (_LINUX_DEBUG) )
extern int rttWaitLWPs();
#endif

/************************************************************
 *  Defines
 ************************************************************/

#define PID_FILENAME_TPL "rtt-test-case.%s.pid"

#define RTT_GET_TIC (rttGetTimeTick(RTTgs.timers_in_microseconds  ? \
			            rttTimeScale_micro_seconds  :   \
			            rttTimeScale_milli_seconds))

/************************************************************
 *  Global variables
 ************************************************************/

static int32_t rttContext_timeGeneration  = 0;

static char *prog_name = NULL;

/*--- identification of this RT-Tester Process ---*/
//static int32_t rttContext_rttProcId       = 0;

// ///////////////////////////////////////////////////////////////////
// Signal Handler
// ///////////////////////////////////////////////////////////////////

/**
 * Function rttSignalHandler.
 * This function defines the actions for the different signals that can be
 * sent to the test process.
 *
 * @TODO missing documentation of parameters
 * @param sig
 */
void rttSignalHandler( int sig ) {
    uint32_t am_id;
    char *tmp;
    rttTimeTic_t cnode_timetic;
#ifdef RTT_CLUSTER
    uint32_t rtt_local_cnode_id = getClusterNode();
#else
    uint32_t rtt_local_cnode_id = 0;
#endif

    cnode_timetic = RTT_GET_TIC - rttStartOfTestTic;

    signal(sig, &rttSignalHandler);

    switch (sig ) {

	case 10: 
	    rttContext_timeGeneration = 1;
#if (VERBOSE_LEVEL > 0) 
	    printf("TM %011lld CN%4d O Received signal 10 - starting test.\n",
		   (long long int)cnode_timetic, rtt_local_cnode_id);
#endif
	    break;

#if (! defined _CADUL) && (! defined _MINGW)
	case SIGPIPE:
	case SIGALRM:
	case SIGQUIT:
	case SIGHUP:
#endif
	case SIGABRT:
	case SIGTERM:
	case SIGINT:
	    if ( rttContext_rttTestRunning ) {
		// stop test execution
		rttContext_rttTestRunning = 0;

		// print output message
		fprintf(stderr,
		       "TM %011lld CN%4lu O Caught termination signal %d, "
		       "stop test execution.\n",
		       (long long int)cnode_timetic, (long unsigned int)rtt_local_cnode_id, sig);
                fflush(0);

		// output message to log of current AM.
		tmp = (char*)malloc(1000);
		
		for(am_id = 0; am_id < RTT_CONFIG_NUM_AM; am_id++) {
		    if (AM_context[am_id].used) {
			sprintf(tmp,
				"ABORTED IN TEST STEP : %ld\n"
				"CAUSE                : CAUGHT SIGNAL %d",
				AM_context[am_id].test_step,
				sig);
			rttctx_set_verdict_AM(am_id, RTTverdict_error, tmp);
			fflush(0);
		    }
		}
	    }
	    break;

	case SIGILL:
	case SIGFPE:
	case SIGSEGV:
	    // stop test execution
	    rttContext_rttTestRunning = 0;

	    fprintf(stderr,
		   "TM %011lld CN%4lu O Exiting on signal %d.\n",
		   (long long int)cnode_timetic, (long unsigned int)rtt_local_cnode_id, sig);
	    fflush(0);

#ifdef RTT_CLUSTER
	    rttClusterdTestStop();
#endif 
	    // exit
	    exit(sig);
	    break;

	default:
	    if ( rttContext_rttTestRunning ) {
		// stop test execution
		rttContext_rttTestRunning = 0;

		// print output message
		fprintf(stderr,
		       "TM %011lld CN%4lu O Caught unexpected signal %d, exiting!\n",
		       (long long int)cnode_timetic, (long unsigned int)rtt_local_cnode_id, sig);
                fflush(0);

#ifdef RTT_CLUSTER
		rttClusterdTestStop();
#endif 
		// exit
		exit(sig);
	    }
	    break;

    }

}


/**
 * Signal handler that prints a backtrace, implemented in RTTschedule.c
 *
 * @TODO missing documentation
 *
 * @param sig
 * @param si
 * @param ctx
 */
void lwpBacktraceSignalHandler(int sig
#ifdef _LINUX
			       , siginfo_t *si
			       , void *ctx
#endif
			       );


#if (! defined _CADUL) && (! defined _MINGW)
/**
 * Function rtt_nanosleep.
 * This function is used during the test run to suspend the control process.
 */
void rtt_nanosleep() {
    int    dummy;
    struct timespec dur, rem;
    uint64_t duration = RTT_CONFIG_GLOBAL_CYCLETIME;
    uint64_t timescale = RTTgs.timers_in_microseconds  ? 1000000  : 1000;

    dur.tv_sec = (time_t)(duration/timescale);
    dur.tv_nsec = (long)((1000000000/timescale)*(duration%timescale));

    dummy = nanosleep(&dur,&rem);
    return;
}
#endif


#if (! defined _CADUL) && (! defined _MINGW)
/**
 * Function rttGetHostName()
 *
 * This function returns the hostname of the cluster node the test is
 * currently running. If the hostname cannot be determined the test is stopped
 * automatically.
 */
static const char *rttGetHostName() {
    static char hostname[1024];
    int ret = gethostname(hostname, 1024);
    if (ret == -1) {
	fprintf(stderr, "Error: Cannot determine hostname.");
#ifdef RTT_CLUSTER
	rttClusterdTestStop();
#endif 
	exit(10);
    }
    return hostname;
}
#endif

/**
 * Functiom createPidReferenceFile.  
 *
 * This function creates a file "rtt-test-case.pid" in the current directory,
 * which contains the PID of the process. This PID can be used to terminate
 * the process manually.
 */
void createPidReferenceFile() {
#if (! defined _CADUL) && (! defined _MINGW)
    FILE              *pidfile;
    char PID_FILENAME[1024];

    sprintf(PID_FILENAME, PID_FILENAME_TPL, rttGetHostName());
    pidfile = fopen(PID_FILENAME, "wb");
    if (pidfile != NULL) {
	fprintf(pidfile, "%d", getpid());
	fclose(pidfile);
    } else {
	fprintf(stderr, "Cannot create PID file \"%s\" in current directory.", 
		PID_FILENAME);
	exit(10);
    }
#endif
#ifdef RTT_CLUSTER
    rttClusterdTestStart();
#endif 
}

/**
 * Functiom removePidReferenceFile.  
 *
 * @note this function could return the value of 'unlink', but does not
 *
 * This function removes a file "rtt-test-case.pid" in the current directory,
 * which contains the PID of the process. 
 */
void removePidReferenceFile() {
    int retval = 0;
#if (! defined _CADUL) && (! defined _MINGW)
    char PID_FILENAME[1024];

    sprintf(PID_FILENAME, PID_FILENAME_TPL, rttGetHostName());
    retval = unlink(PID_FILENAME);
#endif
#ifdef RTT_CLUSTER
    rttClusterdTestStop();
#endif 
    // -- this is to avoid compiler warnings: force variables to be used --
    if(retval < rttContext_timeGeneration){ 
      return;
    }
}

#if (defined _CADUL) || (defined _MINGW)
/** Constant Implementation: Always return 0 */
uint32_t getClusterNode() { 
  return 0; 
}
#else
/**
 * Function getClusterNode.  
 *
 * This function returns the cluster node id of the cluster node, the test is
 * started on. The cluster node id can be used to get the CNODE_context entry
 * of this cluster node. The CNODE_context contains definitions about the LWPs
 * that are running on this node.
 * 
 * @return Cluster node id of the cluster node. Its range is in
 *         0..RTT_CONFIG_NUM_CNODE-1. The id is an index so that
 *         CNODE_context[id] contains the cluster node attributes
 *         of the local node where getClusterNode() has been called.
 */
uint32_t getClusterNode() {
    char hostname[1024];
    uint32_t cnode_idx;
    int ret = gethostname(hostname, 1024);
    if (ret == -1) {
	fprintf(stderr,
		"Error: Cannot determine hostname."
		" (needed for cluster node id)\n");
#ifdef RTT_CLUSTER
	rttClusterdTestStop();
#endif 
	exit(10);
    }
    for (cnode_idx = 0; cnode_idx < RTT_CONFIG_NUM_CNODE; cnode_idx++) {
	if (strcmp(hostname,CNODE_context[cnode_idx].hostname) == 0) {
#if (VERBOSE_LEVEL > 0)
	    fprintf(stdout,
		    "cluster node '%s' has id %d.\n",
		    hostname,cnode_idx);
#endif
	    return cnode_idx;
	}
    }
    fprintf(stderr,
	    "Error: Cluster node '%s' is not defined "
	    "in the test configuration.\n",hostname);
#ifdef RTT_CLUSTER
    rttClusterdTestStop();
#endif 
    exit(10);
}
#endif

// ///////////////////////////////////////////////////////////////////
// parse parameters
// ///////////////////////////////////////////////////////////////////
/**
 * Function parseCommandline.
 * The only allowed commandline so far is an optional max. duration time for
 * the test process.
 * @param argc the number of arguments, including the name of the executable
 * @param argv an array of char*, pointing to the different argument strings
 */
static void  parseCommandline(int argc, char *argv[]) {
    long int parsed_val ; // parsed integer value

#ifndef _CADUL
    prog_name = strdup(argv[0]);
#else
    prog_name = "rtt-test-case.exe";
#endif
    
    if ( argc >= 2 ) {
	if (strcmp(argv[1],"-h") == 0) {
	    fprintf(stderr,
		    "Usage: %s [<TEST-DURATION(SEC)>]\n",
		    prog_name);
#ifdef RTT_CLUSTER
	    rttClusterdTestStop();
#endif 
	    exit(10);
	}
	parsed_val = atol(argv[1]);
	if(parsed_val >= 0L){
	  rttTestDurationSec = (uint64_t)parsed_val;
	}
	else {
	  rttTestDurationSec = 0LL;
	}
    } else {
	// no argument given => endless duration time
	rttTestDurationSec = 0LL;
    }

}


/**
 * rttAllCnodesReady()
 *
 * @return TRUE iff all cluster nodes are in state readyToRun.
 */
#ifdef RTT_CLUSTER
static bool_t rttAllCnodesReady(void) {

    uint32_t cnodeId;

    for (cnodeId = 0; cnodeId < rtt_config_num_cnode; cnodeId++) {
	if ( ! CNODE_context[cnodeId].readyToRun ) return FALSE;
    }

    return TRUE;

}
#endif

// ///////////////////////////////////////////////////////////////////
// MAIN
// ///////////////////////////////////////////////////////////////////


/**
   The main routine starts up the unit test.

   The following signals are caught and yield a test termination
   with finializing log entry:

   <PRE> 
   Signals described in the original POSIX.1 standard.


   Signal     Value     Action   Comment
   -------------------------------------------------------------------------
   SIGHUP        1       Term    Hangup detected on controlling terminal
   or death of controlling process
   SIGINT        2       Term    Interrupt from keyboard
   SIGQUIT       3       Core    Quit from keyboard
   SIGILL        4       Core    Illegal Instruction
   SIGABRT       6       Core    Abort signal from abort(3)
   SIGFPE        8       Core    Floating point exception
   SIGKILL       9       Term    Kill signal
   SIGSEGV      11       Core    Invalid memory reference
   SIGPIPE      13       Term    Broken pipe: write to pipe with no readers
   SIGALRM      14       Term    Timer signal from alarm(2)
   SIGTERM      15       Term    Termination signal
   SIGUSR1   30,10,16    Term    User-defined signal 1
   SIGUSR2   31,12,17    Term    User-defined signal 2
   SIGCHLD   20,17,18    Ign     Child stopped or terminated
   SIGCONT   19,18,25            Continue if stopped
   SIGSTOP   17,19,23    Stop    Stop process
   SIGTSTP   18,20,24    Stop    Stop typed at tty
   SIGTTIN   21,21,26    Stop    tty input for background process
   SIGTTOU   22,22,27    Stop    tty output for background process

   The signals SIGKILL and SIGSTOP cannot be caught, blocked, or ignored.
   </PRE>
*/

/** ensure that stdout and stderr are opened in binary mode */
RTT_INCLUDE_THIS_LINE_BEFORE_MAIN;


/**
 * Function main.
 * The main function of each test case. This function controls the startup and
 * the test termination.
 * @param argc the number of arguments, including the name of the executable
 * @param argv an array of char*, pointing to the different argument strings
 */
int main (int argc, char *argv[]) {
    time_t t_val;
    rtt_lwp_status_t status_of_lwps = rtt_lwp_initialising;

#if ( (defined (_LINUX) || defined (RTT_PTHREAD) ) && ! defined (_LINUX_DEBUG)  )
    uint32_t lwp_id, lwp_idx;
    uint32_t cnode_id;
#ifdef RTT_CLUSTER
    rttTimeTic_t cnode_timetic;
#endif
#endif
#ifdef _LINUX
    struct sigaction sa;

    {
      // by default, block all user-defined signals;
      // this will be inherited by all spawned pthreads (if used)
      sigset_t sigset;
      int signo;
      sigemptyset(&sigset);
      for (signo = SIGRTMIN; signo <= SIGRTMAX; ++signo) {
        sigaddset(&sigset, signo);
      }
#ifdef RTT_PTHREAD
      pthread_sigmask( SIG_BLOCK, &sigset, NULL );
#else
      sigprocmask( SIG_BLOCK, &sigset, NULL );
#endif
    }
#endif

    // load symbols for backtrace
    rttctx_init_symbols();

    // parse commandline parameters
    parseCommandline(argc, argv);

    // Get the thread ID for the control thread
    rttControlThreadId = RTT_GETTID();

    // initialise Context for 
    // stub test integration levels, AMs, LWPs and channels
    createPidReferenceFile();

    am_context_INIT();
    lwp_context_INIT();
    cnode_context_INIT();
    channel_context_INIT();
    vector_context_INIT();
    stublevel_context_INIT();

    // register signal handlers
    if (!rttIgnoreSignal_SIGINT)  signal(SIGINT,   &rttSignalHandler);
    if (!rttIgnoreSignal_SIGTERM) signal(SIGTERM,  &rttSignalHandler);
#if (! defined _CADUL) && (! defined _MINGW)
    if (!rttIgnoreSignal_SIGHUP)  signal(SIGHUP,   &rttSignalHandler);
    if (!rttIgnoreSignal_SIGUSR1) signal(SIGUSR1,  &rttSignalHandler);
    if (!rttIgnoreSignal_SIGUSR2) signal(SIGUSR2,  &rttSignalHandler);
    if (!rttIgnoreSignal_SIGCHLD) signal(SIGCHLD,  &rttSignalHandler);
    if (!rttIgnoreSignal_SIGCONT) signal(SIGCONT,  &rttSignalHandler);
    if (!rttIgnoreSignal_SIGSTOP) signal(SIGSTOP,  &rttSignalHandler);
    if (!rttIgnoreSignal_SIGTSTP) signal(SIGTSTP,  &rttSignalHandler);
    if (!rttIgnoreSignal_SIGTTIN) signal(SIGTTIN,  &rttSignalHandler);
    if (!rttIgnoreSignal_SIGTTOU) signal(SIGTTOU,  &rttSignalHandler);
#endif

    // exit on error signals
#ifdef _LINUX
    /* use 3-parameter signal handler on linux */
    sa.sa_sigaction = lwpBacktraceSignalHandler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_SIGINFO;

    if (!rttIgnoreSignal_SIGILL)  sigaction(SIGILL,  &sa, NULL);
    if (!rttIgnoreSignal_SIGABRT) sigaction(SIGABRT, &sa, NULL);
    if (!rttIgnoreSignal_SIGFPE)  sigaction(SIGFPE,  &sa, NULL);
    if (!rttIgnoreSignal_SIGPIPE) sigaction(SIGPIPE, &sa, NULL);
    if (!rttIgnoreSignal_SIGALRM) sigaction(SIGALRM, &sa, NULL);
    if (!rttIgnoreSignal_SIGQUIT) sigaction(SIGQUIT, &sa, NULL);
#if ( ! defined (_LINUX_DEBUG) )
    if (!rttIgnoreSignal_SIGSEGV) sigaction(SIGSEGV, &sa, NULL);
#endif
#else
    if (!rttIgnoreSignal_SIGILL)  signal(SIGILL,  lwpBacktraceSignalHandler);
    if (!rttIgnoreSignal_SIGABRT) signal(SIGABRT, lwpBacktraceSignalHandler);
    if (!rttIgnoreSignal_SIGFPE)  signal(SIGFPE,  lwpBacktraceSignalHandler);
    if (!rttIgnoreSignal_SIGSEGV) signal(SIGSEGV, lwpBacktraceSignalHandler);
#if (! defined _CADUL) && (! defined _MINGW)
    if (!rttIgnoreSignal_SIGPIPE) signal(SIGPIPE, lwpBacktraceSignalHandler);
    if (!rttIgnoreSignal_SIGQUIT) signal(SIGQUIT, lwpBacktraceSignalHandler);
    if (!rttIgnoreSignal_SIGALRM) signal(SIGALRM, lwpBacktraceSignalHandler);
#endif
#endif

    // switch to initialisation phase
    rttContext_rttTestRunning = 1;
    status_of_lwps = rtt_lwp_initialising;

    // calculate global start time and epoch

    rttStartOfTestTic = rttGetTimeTick(RTTgs.timers_in_microseconds  ? 
				       rttTimeScale_micro_seconds  :
				       rttTimeScale_milli_seconds);
    t_val = time((time_t *)&rttStartOfTestEpoch);
    if ( t_val == ((time_t)-1) )
	perror("error calculating absolute startup time:");


#ifdef RTT_CLUSTER
    rtt_local_cnode_id = getClusterNode();
    // setup Myrinet or TCP/IP communications with other nodes
    rttClusterInit(rtt_local_cnode_id, RTT_CONFIG_NUM_CNODE);
#endif

    // create test LWPs for AMs
    rttCreateLWPs();


#if ( (defined (_LINUX) || defined (RTT_PTHREAD) ) && ! defined (_LINUX_DEBUG)  )

    // check for cluster node configuration
    if (RTT_CONFIG_NUM_CNODE > 0) {
	// calculate cluster node id
	cnode_id = getClusterNode();
	// wait for LWPs to be ready to run
	while (status_of_lwps == rtt_lwp_initialising) {
	    status_of_lwps = rtt_lwp_ready_to_run;
	    for (lwp_idx=0; 
		 lwp_idx<(uint32_t)CNODE_context[cnode_id].num_of_lwps;
		 lwp_idx++) {
		lwp_id = CNODE_context[cnode_id].lwp_id_list[lwp_idx];
		if (LWP_context[lwp_id].lwp_status == rtt_lwp_initialising)
		    status_of_lwps = rtt_lwp_initialising;
	    }
	}
    } else {
	// wait for LWPs to be ready to run
	while (status_of_lwps == rtt_lwp_initialising) {
	    status_of_lwps = rtt_lwp_ready_to_run;
	    for (lwp_id=0; lwp_id<RTT_CONFIG_NUM_LWP; lwp_id++) {
		if (LWP_context[lwp_id].lwp_status == rtt_lwp_initialising)
		    status_of_lwps = rtt_lwp_initialising;
	    }
	}
    }

    if (rttContext_rttTestRunning != 0) {
      // create map of channels which are written to / read from 
      // on specific cluster nodes
      rttCreateChannelMap();
    }
    
    // startup test now, unless @INIT already perfored '@rttStopTest'
    if (rttContext_rttTestRunning != 0) {

#if (VERBOSE_LEVEL > 0)
	cnode_timetic = RTT_GET_TIC - rttStartOfTestTic;
#ifdef RTT_CLUSTER
	printf("TM %011lld CN%4d O Starting test on cluster node.\n",
	       cnode_timetic, rtt_local_cnode_id);
#else
	printf("TM %011lld CN%4d O Starting test on cluster node.\n",
	       cnode_timetic, 0);
#endif
	fflush(NULL);
#endif

	/** Cluster startup synchronisation */
#ifdef RTT_CLUSTER
	{
	  uint32_t remoteCnodeId;
	  uint32_t cmd;
	  uint32_t cn;
	  uint32_t cnlId = 0;
	    
	  /*--- send my channel subscriptions to other nodes ---*/
	  for ( cnlId = 0; cnlId < RTT_NUM_CHANNEL_MAP; cnlId++) {

	    if ( rttReadChannelMap[cnlId] )
	      for ( cn = 0; cn < rtt_config_num_cnode; cn++) {
		if ( cn != rtt_local_cnode_id )
		  rttcluster_tx_subscribecnlcmd(cn,cnlId);
		rttRxClusterCmd(&cmd,&remoteCnodeId);
		if (cmd == RTT_CLUSTER_INITIALISED_TEST) {
		  CNODE_context[remoteCnodeId].readyToRun = TRUE;   
		  cnode_timetic = RTT_GET_TIC - rttStartOfTestTic;
		  printf("TM %011lld CN%4d O RTT_CLUSTER_INITIALISED_TEST FROM CN %d\n",
			 cnode_timetic, rtt_local_cnode_id, remoteCnodeId);
		  fflush(NULL);
		}
	      }
	  }
		
	  /*--- Send my "end of initialisation" message to all nodes ---*/
	  rttTxClusterCmd(RTT_CLUSTER_INITIALISED_TEST);
	  cnode_timetic = RTT_GET_TIC - rttStartOfTestTic;
	  printf("TM %011lld CN%4d O RTT_CLUSTER_INITIALISED_TEST\n",
		 cnode_timetic, rtt_local_cnode_id);
	  fflush(NULL);
	  CNODE_context[rtt_local_cnode_id].readyToRun = TRUE;

	  do {

	    rttRxClusterCmd(&cmd,&remoteCnodeId);
	    if (cmd == RTT_CLUSTER_INITIALISED_TEST) {
	      CNODE_context[remoteCnodeId].readyToRun = TRUE;   
	      cnode_timetic = RTT_GET_TIC - rttStartOfTestTic;
	      printf("TM %011lld CN%4d O RTT_CLUSTER_INITIALISED_TEST FROM CN %d\n",
		     cnode_timetic, rtt_local_cnode_id, remoteCnodeId);
	    }

	  } while ( ! rttAllCnodesReady() );

	}
    
	printf("TM %011lld CN%4d O Cluster node initialisation completed\n",
	       cnode_timetic, rtt_local_cnode_id);
	fflush(NULL);
	
#endif
	rttContext_rttTestRunning = 2;
    }

    status_of_lwps = rtt_lwp_running;
    // check for cluster node configuration
    // @todo remove this busy wait stuff - 
    // instead wait for all LWPs to terminate
    if (RTT_CONFIG_NUM_CNODE > 0) {
	// check for running LWPs
	while ( rttContext_rttTestRunning &&
	        (status_of_lwps == rtt_lwp_running)) {
	    status_of_lwps = rtt_lwp_test_completed;
	    for (lwp_idx=0; 
		 lwp_idx<(uint32_t)CNODE_context[cnode_id].num_of_lwps;
		 lwp_idx++) {
		lwp_id = CNODE_context[cnode_id].lwp_id_list[lwp_idx];
		if ((LWP_context[lwp_id].lwp_status == rtt_lwp_ready_to_run) ||
		    (LWP_context[lwp_id].lwp_status == rtt_lwp_running))
		    status_of_lwps = rtt_lwp_running;
	    }
	    // wait for LWP
	    rtt_nanosleep();
	}
    } else {
	// check for running LWPs
	while (rttContext_rttTestRunning &&
	       (status_of_lwps == rtt_lwp_running)) {
	    status_of_lwps = rtt_lwp_test_completed;
	    for (lwp_id=0; lwp_id<RTT_CONFIG_NUM_LWP; lwp_id++) {
		if ((LWP_context[lwp_id].lwp_status == rtt_lwp_ready_to_run) ||
		    (LWP_context[lwp_id].lwp_status == rtt_lwp_running))
		    status_of_lwps = rtt_lwp_running;
	    }
	    // wait for LWP
	    rtt_nanosleep();
	}
    }

    // set global test termination flag
    rttContext_rttTestRunning = 0;

    // wait for child processes to terminate
#ifndef _LINUX_DEBUG
    rttWaitLWPs();
#endif
#endif

    am_context_FINIT();
    lwp_context_FINIT();
    cnode_context_FINIT();
    channel_context_FINIT();
    vector_context_FINIT();
    stublevel_context_FINIT();

    removePidReferenceFile();

#ifdef RTT_CLUSTER
    rttClusterFinit();
#endif 

#ifdef _CADUL
    // -- this is to avoid compiler warnings ----------------------------
    if(rtt_lwp_initialising == status_of_lwps){
      return 0;
    }
#endif

    return 0;
}


/*//////////////////////////////////////////////////////////////
//
// Change History:
//
// $Log: RTTstartup.c,v $
// Revision 1.47  2007/08/21 08:02:04  oli
// added macro RTT_INCLUDE_THIS_LINE_BEFORE_MAIN:
// ensure that RT-Tester executables open stdout, stderr as binary streams
// (no DOS line breaks in MinGW build)
//
// Revision 1.46  2007/08/06 14:10:22  oli
// prepared introduction of (compiler) configuration field FAMILY
// new: mingw family
// prepared build on Win32/Cygwin that compiles executables with mingw
// (without reference to cygwin1.dll)
//
// Revision 1.45  2007/05/29 08:09:59  oli
// RTTstartup.c: added uint32_t getClusterNode(){ return 0; }
// this fixes (rt-tester/240)
//
// Revision 1.44  2007/04/26 07:00:03  oli
// added (silly, but harmless) code in order to avoid Cadul compiler warnings
//
// Revision 1.43  2007/04/20 08:54:10  oli
// use type cast to avoid compiler warnings (g++ (GCC) 4.1.2 20061115)
//
// Revision 1.42  2007/04/19 11:53:16  oli
// added int casts in order to avoid compile warnings
//
// Revision 1.41  2007/04/19 11:50:56  oli
// robustness: do not declare variable inside for() construct
// (not allowed with gcc (GCC) 4.0.2 20050901)
// added int casts in order to avoid compile warnings
//
// Revision 1.40  2007/04/19 09:32:05  jp
// Mask RT-Signals SIGRTMIN..SIGRTMAX for RT-Tester
//
// Revision 1.39  2007/04/18 12:09:53  oli
// removed compile warnings by int casts for format strings;
// declare rttWaitLWPs, if appropriate
//
// Revision 1.38  2007/03/20 09:18:56  oli
// robustness: also check for rttContext_rttTestRunning != 0
// after call to rttCreateChannelMap() (rt-tester/220,223)
//
// Revision 1.37  2007/03/19 18:03:52  oli
// fixed @init:@rttstopTest (rt-tester/220):
//
// If after the @INIT phase, rttContext_rttTestRunning == 0,
// then do not call rttCreateChannelMap(), for it might crash.
//
// Revision 1.36  2007/01/11 15:03:28  hjficker
// Fixed compiler warnings
//
// Revision 1.35  2007/01/10 14:55:50  hjficker
// Install lwpBacktraceSignalHandler() for error signals also for the
// control thread.
//
// Revision 1.34  2007/01/10 13:54:43  hjficker
// Fix for RTT problem report 207: When the lwpSignalHandler is called from
// the control thread, it calls rttSignalHandler instead of executing the LWP
// related signal handling code.
//
// Revision 1.33  2006/12/13 17:29:33  dahlweid
// Fixed initialisation of LWPs in pthread configuration
//
// Revision 1.32  2006/10/13 15:33:32  chref
// adjustments necessary for dynamic AM handling:
// - always have 512 entries in AM_context
// - used-flag in AM_context entries (in order to find free entries)
// - always have 512 entries in am_id_list, am_start_list and am_prio_list
// - keep "volatile" attribute when passing pointers to LWP_context entries
//   to functions
//
// Revision 1.31  2006/08/08 11:07:40  chref
// load symbols for backtrace on test startup, because loading can fail
// if performed when memory/stack has already been corrupted
//
// Revision 1.30  2006/07/03 13:14:34  oli
// fixed: faulty uses of printf(stderr, "..." ) --> fprintf(stderr, "...")
//
// Revision 1.29  2006/06/27 07:57:49  dahlweid
// Logging to console is now generated in the RT-Tester typical timestamp format.
//
// Revision 1.28  2006/03/03 16:27:23  oli
// fixed: "gethostname" for cygwin (disabled for _CADUL)
//
// Revision 1.27  2006/02/23 09:21:49  dahlweid
// Replaced rtt-test-case.pid by rtt-test-case.<hostname>.pid
//
// Revision 1.26  2006/01/16 15:51:05  oli
// replaced all occurrences of fopen: "w", "r":
//    fopen(..., "r")   --> fopen(..., "rb")
//    fopen(..., "w")   --> fopen(..., "wb")
//
// Revision 1.25  2005/11/09 15:54:44  dahlweid
// - Added variables rttIgnoreSignal_<signal>, which are initialised from
//   RTTcontext.c and allow to disable signal handlers for the specified signals
//   with compile switches for each test procedure.
// - Fixed C++ compile problem of rttHaveDebugMode.
//
// Revision 1.24  2005/09/21 14:47:25  chref
// marked global variables that get changed in different threads as "volatile"
//
// Revision 1.23  2005/08/18 19:37:55  walsen
// make rtt-test-case templates compile with cadul
//
// Revision 1.22  2005/08/12 15:55:02  walsen
// further cad-ul adjustments
//
// Revision 1.21  2005/08/10 06:59:29  jp
// First version of the TCP/IP cluster communication,
// as an configurable alternative. Not yet tested, but note that
// changes will only become visible if you configure with
//     --enable-tcpcluster
//
// Revision 1.20  2005/07/14 16:17:18  jp
// Completion of startup protocol with channel subscription in cluster
//
// Revision 1.19  2005/06/24 08:44:01  dahlweid
// Replaced mring references by generic cluster function calls
//
// Revision 1.18  2005/05/16 13:45:46  jp
// Typos corrected and comments changed
//
// Revision 1.17  2005/04/08 08:46:28  jp
// Multi-threaded cluster nodes - first extensions
//
// Revision 1.16  2005/02/08 11:22:06  dahlweid
// - Included patches from Stefan Walsen, who ported RT-Tester 6.0 MacOSX
//   (Darwin).
// - Use autoconf macros for pth checks.
//
// Revision 1.15  2004/11/18 08:50:04  dahlweid
// Support Verified Start-Stop-Controller from RT-Tester 6.x. On test start the
// clusterd is informed about test startup; on test completion the flashing
// indication is reset via clusterd.
//
// Revision 1.14  2004/08/04 14:53:21  marlex
// functions comments inserted/changed to coding standard
// for more information (and TODO) see in TODO_for_coding_standard.txt
//
// Revision 1.13  2004/08/02 14:43:01  marlex
// header added/changed for coding standard in *.l, *.y, *.c, *.h files
//
//
//
///////////////////////////////////////////////////////////////
*/
