#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <assert.h>

#include "interface.h"


typedef enum { s_q1, s_q0, s_p } state_t;


void startClock(struct timeval *tv)
{
    gettimeofday(tv, NULL);
    tv->tv_usec /= (1000000 / TIME_UNIT);
}


int getClock(const struct timeval *tv)
{
    struct timeval now, diff;
    int result;

    startClock(&now);
    timersub(&now, tv, &diff);

    result = diff.tv_sec * TIME_UNIT + diff.tv_usec;
    return result;
}


void ta(int cfd)
{
    state_t state;
    data_t data;
    struct timeval x, y;

    state = s_q1;
    startClock(&x);
    startClock(&y);

    for (;;) {
	switch (state) {
	case s_q1:
	    assert(getClock(&y) <= 1 * TIME_UNIT);

	    if (getClock(&x) == 0 * TIME_UNIT &&
		getClock(&y) < 1 * TIME_UNIT) {
		if (read(cfd, &data, sizeof data) == sizeof data &&
		    data == d_in) {
		    state = s_p;
		    continue;
		}
	    }

	    if (getClock(&x) > 0 * TIME_UNIT &&
		getClock(&y) < 1 * TIME_UNIT) {
		if (read(cfd, &data, sizeof data) == sizeof data &&
		    data == d_in) {
		    state = s_q0;
		    startClock(&x);
		    continue;
		}
	    }

	    if (getClock(&y) == 1 * TIME_UNIT) {
		data = d_out;
		write(cfd, &data, sizeof data);
		state = s_p;
		continue;
	    }
	    break;
	    
	case s_q0:
	    assert(getClock(&y) <= 1 * TIME_UNIT);

	    if (getClock(&y) < 0 * TIME_UNIT) {
		if (read(cfd, &data, sizeof data) == sizeof data &&
		    data == d_in) {
		    state = s_p;
		    continue;
		}
	    }

	    if (getClock(&y) == 1 * TIME_UNIT) {
		data = d_out;
		write(cfd, &data, sizeof data);
		state = s_p;
	    }
	    break;
	    
	case s_p:
	    assert(1);

	    if (read(cfd, &data, sizeof data) == sizeof data &&
		data == d_in) {
		data = d_out;
		write(cfd, &data, sizeof data);
		state = s_p;
		continue;
	    }
	    break;
	}
    }
}


int main(int argc, char *argv[])
{
    int sfd, cfd;
    sfd = socket(PF_INET, SOCK_STREAM, 0);

    if (sfd != -1) {
	struct sockaddr_in si = { AF_INET,
				  ntohs(INTERFACE_PORT),
				  { ntohl(INADDR_LOOPBACK) } };

	if (bind(sfd, (struct sockaddr *)&si, sizeof si) == 0) {
	    if (listen(sfd, 1) == 0) {
		cfd = accept(sfd, NULL, NULL);
		if (cfd != -1) {
		    data_t data;

		    if (read(cfd, &data, sizeof data) == sizeof data &&
			data == d_start) {
			fcntl(cfd, F_SETFL, fcntl(cfd, F_GETFL) | O_NONBLOCK);
			ta(cfd);
		    }
		    close(cfd);
		}
		else {
		    perror("accept");
		}
	    }
	    else {
		perror("listen");
	    }
	}
	else {
	    perror("bind");
	}

	close(sfd);
    }
    else {
	perror("socket");
    }

    return EXIT_SUCCESS;
}
